@@ -1729,6 +1729,10 @@ static gboolean migrate_connect(gpointer data)
                          "verify", SPICE_SESSION_VERIFY_PUBKEY,
                          NULL);
             g_byte_array_unref(pubkey);
+        } else if (info->cert_subject_size == 0 ||
+                   strlen((const char*)info->cert_subject_data) == 0) {
+            /* only verify hostname if no cert subject */
+            g_object_set(mig->session, "verify", SPICE_SESSION_VERIFY_HOSTNAME, NULL);
         } else {
             gchar *subject = g_alloca(info->cert_subject_size + 1);
             strncpy(subject, (const char*)info->cert_subject_data, info->cert_subject_size);
@@ -2241,4 +2245,9 @@ void spice_main_set_display_enabled(SpiceMainChannel *channel, int id, gboolean
         g_return_if_fail(id < G_N_ELEMENTS(c->display));
         c->display[id].enabled = enabled;
     }
+
+    if (c->timer_id) {
+        g_source_remove(c->timer_id);
+    }
+    c->timer_id = g_timeout_add_seconds(1, timer_set_display, channel);
 }
