@@ -20,6 +20,8 @@
 #include <glib-object.h>
 #include <string.h>
 
+#include "glib-compat.h"
+
 #ifdef USE_SMARTCARD
 #include <vcard_emul.h>
 #include <vevent.h>
@@ -38,7 +40,7 @@
  * @section_id:
  * @see_also:
  * @stability: Stable
- * @include: spice-smartcard-manager.h
+ * @include: smartcard-manager.h
  *
  * #SpiceSmartcardManager monitors smartcard reader plugging/unplugging,
  * and smartcard insertions/removals. It also provides methods to handle
@@ -50,9 +52,9 @@
 /* gobject glue                                                       */
 
 #define SPICE_SMARTCARD_MANAGER_GET_PRIVATE(obj)                                  \
-    (G_TYPE_INSTANCE_GET_PRIVATE ((obj), SPICE_TYPE_SMARTCARD_MANAGER, spice_smartcard_manager))
+    (G_TYPE_INSTANCE_GET_PRIVATE ((obj), SPICE_TYPE_SMARTCARD_MANAGER, SpiceSmartcardManagerPrivate))
 
-struct spice_smartcard_manager {
+struct _SpiceSmartcardManagerPrivate {
     guint monitor_id;
 
     /* software smartcard reader, the certificates to use for this reader
@@ -69,7 +71,8 @@ G_DEFINE_TYPE(SpiceSmartcardManager, spice_smartcard_manager, G_TYPE_OBJECT)
 #ifdef USE_SMARTCARD
 G_DEFINE_BOXED_TYPE(VReader, spice_smartcard_reader, vreader_reference, vreader_free)
 #else
-G_DEFINE_BOXED_TYPE(GObject, spice_smartcard_reader, NULL, NULL)
+typedef GObject VReader;
+G_DEFINE_BOXED_TYPE(VReader, spice_smartcard_reader, g_object_ref, g_object_unref)
 #endif
 
 /* Properties */
@@ -100,7 +103,7 @@ static gboolean smartcard_monitor_dispatch(VEvent *event, gpointer user_data);
 
 static void spice_smartcard_manager_init(SpiceSmartcardManager *smartcard_manager)
 {
-    spice_smartcard_manager *priv;
+    SpiceSmartcardManagerPrivate *priv;
 
     priv = SPICE_SMARTCARD_MANAGER_GET_PRIVATE(smartcard_manager);
     smartcard_manager->priv = priv;
@@ -119,7 +122,7 @@ static void spice_smartcard_manager_dispose(GObject *gobject)
 
 static void spice_smartcard_manager_finalize(GObject *gobject)
 {
-    spice_smartcard_manager *priv;
+    SpiceSmartcardManagerPrivate *priv;
 
     priv = SPICE_SMARTCARD_MANAGER_GET_PRIVATE(gobject);
     if (priv->monitor_id != 0) {
@@ -157,7 +160,7 @@ static void spice_smartcard_manager_class_init(SpiceSmartcardManagerClass *klass
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET(SpiceSmartcardManagerClass, reader_added),
                      NULL, NULL,
-                     g_cclosure_user_marshal_VOID__BOXED,
+                     g_cclosure_marshal_VOID__BOXED,
                      G_TYPE_NONE,
                      1,
                      SPICE_TYPE_SMARTCARD_READER);
@@ -176,7 +179,7 @@ static void spice_smartcard_manager_class_init(SpiceSmartcardManagerClass *klass
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET(SpiceSmartcardManagerClass, reader_removed),
                      NULL, NULL,
-                     g_cclosure_user_marshal_VOID__BOXED,
+                     g_cclosure_marshal_VOID__BOXED,
                      G_TYPE_NONE,
                      1,
                      SPICE_TYPE_SMARTCARD_READER);
@@ -196,7 +199,7 @@ static void spice_smartcard_manager_class_init(SpiceSmartcardManagerClass *klass
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET(SpiceSmartcardManagerClass, card_inserted),
                      NULL, NULL,
-                     g_cclosure_user_marshal_VOID__BOXED,
+                     g_cclosure_marshal_VOID__BOXED,
                      G_TYPE_NONE,
                      1,
                      SPICE_TYPE_SMARTCARD_READER);
@@ -216,14 +219,14 @@ static void spice_smartcard_manager_class_init(SpiceSmartcardManagerClass *klass
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET(SpiceSmartcardManagerClass, card_removed),
                      NULL, NULL,
-                     g_cclosure_user_marshal_VOID__BOXED,
+                     g_cclosure_marshal_VOID__BOXED,
                      G_TYPE_NONE,
                      1,
                      SPICE_TYPE_SMARTCARD_READER);
     gobject_class->dispose      = spice_smartcard_manager_dispose;
     gobject_class->finalize     = spice_smartcard_manager_finalize;
 
-    g_type_class_add_private(klass, sizeof(spice_smartcard_manager));
+    g_type_class_add_private(klass, sizeof(SpiceSmartcardManagerPrivate));
 }
 
 /* ------------------------------------------------------------------ */
@@ -244,7 +247,7 @@ static SpiceSmartcardManager *spice_smartcard_manager_new(void)
  * to it. A new SpiceSmartcardManager instance will be created the first
  * time this function is called
  *
- * Returns: a pointer to the #SpiceSmartcardManager singleton
+ * Returns: (transfer none): a weak reference to the #SpiceSmartcardManager
  */
 SpiceSmartcardManager *spice_smartcard_manager_get(void)
 {
@@ -312,8 +315,6 @@ struct _SmartcardSource {
 };
 typedef struct _SmartcardSource SmartcardSource;
 
-typedef gboolean (*SmartcardSourceFunc)(VEvent *event, gpointer user_data);
-
 static gboolean smartcard_source_prepare(GSource *source, gint *timeout)
 {
     SmartcardSource *smartcard_source = (SmartcardSource *)source;
@@ -531,7 +532,7 @@ gboolean spice_smartcard_manager_init_finish(SpiceSession *session,
  * @manager: a #SpiceSmartcardManager
  *
  * Simulates the insertion of a smartcard in the guest. Valid certificates
- * must have been set in #SpiceSession::smartcard-certificates for software
+ * must have been set in #SpiceSession:smartcard-certificates for software
  * smartcard support to work. At the moment, only one software smartcard
  * reader is supported, that's why there is no parameter to indicate which
  * reader to insert the card in.
