@@ -68,9 +68,6 @@ struct _SpiceMsgIn {
 enum spice_channel_state {
     SPICE_CHANNEL_STATE_UNCONNECTED = 0,
     SPICE_CHANNEL_STATE_CONNECTING,
-    SPICE_CHANNEL_STATE_LINK_HDR,
-    SPICE_CHANNEL_STATE_LINK_MSG,
-    SPICE_CHANNEL_STATE_AUTH,
     SPICE_CHANNEL_STATE_READY,
     SPICE_CHANNEL_STATE_SWITCHING,
     SPICE_CHANNEL_STATE_MIGRATING,
@@ -134,6 +131,7 @@ struct _SpiceChannelPrivate {
 
     gsize                       total_read_bytes;
     uint64_t                    last_message_serial;
+    GSList                      *flushing;
 };
 
 SpiceMsgIn *spice_msg_in_new(SpiceChannel *channel);
