@@ -619,9 +619,9 @@ static void clipboard_get(GtkClipboard *clipboard,
 
     /* apparently, this is needed to avoid dead-lock, from
        gtk_dialog_run */
-    GDK_THREADS_LEAVE();
+    gdk_threads_leave();
     g_main_loop_run(ri.loop);
-    GDK_THREADS_ENTER();
+    gdk_threads_enter();
 
     g_main_loop_unref(ri.loop);
     ri.loop = NULL;
