@@ -119,7 +119,7 @@ enum {
 static guint signals[SPICE_DISPLAY_LAST_SIGNAL];
 
 #ifdef WIN32
-static HWND focus_window = NULL;
+static HWND win32_window = NULL;
 #endif
 
 static void update_keyboard_grab(SpiceDisplay *display);
@@ -482,7 +482,11 @@ static void spice_display_init(SpiceDisplay *display)
                           GDK_LEAVE_NOTIFY_MASK |
                           GDK_KEY_PRESS_MASK |
                           GDK_SCROLL_MASK);
+#ifdef WITH_X11
     gtk_widget_set_double_buffered(widget, false);
+#else
+    gtk_widget_set_double_buffered(widget, true);
+#endif
     gtk_widget_set_can_focus(widget, true);
     gtk_widget_set_has_window(widget, true);
     d->keycode_map = vnc_display_keymap_gdk2xtkbd_table(&d->keycode_maplen);
@@ -571,13 +575,13 @@ void spice_display_set_grab_keys(SpiceDisplay *display, SpiceGrabSequence *seq)
 #ifdef WIN32
 static LRESULT CALLBACK keyboard_hook_cb(int code, WPARAM wparam, LPARAM lparam)
 {
-    if  (focus_window && code == HC_ACTION) {
+    if  (win32_window && code == HC_ACTION) {
         KBDLLHOOKSTRUCT *hooked = (KBDLLHOOKSTRUCT*)lparam;
         DWORD dwmsg = (hooked->flags << 24) | (hooked->scanCode << 16) | 1;
 
         if (hooked->vkCode == VK_NUMLOCK || hooked->vkCode == VK_RSHIFT) {
             dwmsg &= ~(1 << 24);
-            SendMessage(focus_window, wparam, hooked->vkCode, dwmsg);
+            SendMessage(win32_window, wparam, hooked->vkCode, dwmsg);
         }
         switch (hooked->vkCode) {
         case VK_CAPITAL:
@@ -591,7 +595,7 @@ static LRESULT CALLBACK keyboard_hook_cb(int code, WPARAM wparam, LPARAM lparam)
         case VK_RMENU:
             break;
         default:
-            SendMessage(focus_window, wparam, hooked->vkCode, dwmsg);
+            SendMessage(win32_window, wparam, hooked->vkCode, dwmsg);
             return 1;
         }
     }
@@ -689,17 +693,102 @@ static void update_keyboard_grab(SpiceDisplay *display)
         try_keyboard_ungrab(display);
 }
 
+static void set_mouse_accel(SpiceDisplay *display, gboolean enabled)
+{
+    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+
+#if defined GDK_WINDOWING_X11
+    GdkWindow *w = GDK_WINDOW(gtk_widget_get_window(GTK_WIDGET(display)));
+    Display *x_display = GDK_WINDOW_XDISPLAY(w);
+
+    if (enabled) {
+        /* restore mouse acceleration */
+        XChangePointerControl(x_display, True, True,
+                              d->x11_accel_numerator, d->x11_accel_denominator, d->x11_threshold);
+    } else {
+        XGetPointerControl(x_display,
+                           &d->x11_accel_numerator, &d->x11_accel_denominator, &d->x11_threshold);
+        /* set mouse acceleration to default */
+        XChangePointerControl(x_display, True, True, -1, -1, -1);
+        SPICE_DEBUG("disabled X11 mouse motion %d %d %d",
+                    d->x11_accel_numerator, d->x11_accel_denominator, d->x11_threshold);
+    }
+#elif defined GDK_WINDOWING_WIN32
+    if (enabled) {
+        g_return_if_fail(SystemParametersInfo(SPI_SETMOUSE, 0, &d->win_mouse, 0));
+        g_return_if_fail(SystemParametersInfo(SPI_SETMOUSESPEED, 0, (PVOID)d->win_mouse_speed, 0));
+    } else {
+        int accel[3] = { 0, 0, 0 }; // disabled
+        g_return_if_fail(SystemParametersInfo(SPI_GETMOUSE, 0, &d->win_mouse, 0));
+        g_return_if_fail(SystemParametersInfo(SPI_GETMOUSESPEED, 0, &d->win_mouse_speed, 0));
+        g_return_if_fail(SystemParametersInfo(SPI_SETMOUSE, 0, &accel, SPIF_SENDCHANGE));
+        g_return_if_fail(SystemParametersInfo(SPI_SETMOUSESPEED, 0, (PVOID)10, SPIF_SENDCHANGE)); // default
+    }
+#else
+    g_warning("Mouse acceleration code missing for your platform");
+#endif
+}
+
+#ifdef WIN32
+static gboolean win32_clip_cursor(void)
+{
+    RECT window, workarea, rect;
+    HMONITOR monitor;
+    MONITORINFO mi = { 0, };
+
+    g_return_val_if_fail(win32_window != NULL, FALSE);
+
+    if (!GetWindowRect(win32_window, &window))
+        goto error;
+
+    monitor = MonitorFromRect(&window, MONITOR_DEFAULTTONEAREST);
+    g_return_val_if_fail(monitor != NULL, false);
+
+    mi.cbSize = sizeof(mi);
+    if (!GetMonitorInfo(monitor, &mi))
+        goto error;
+    workarea = mi.rcWork;
+
+    if (!IntersectRect(&rect, &window, &workarea)) {
+        g_critical("error clipping cursor");
+        return false;
+    }
+
+    SPICE_DEBUG("clip rect %ld %ld %ld %ld\n",
+                rect.left, rect.right, rect.top, rect.bottom);
+
+    if (!ClipCursor(&rect))
+        goto error;
+
+    return true;
+
+error:
+    {
+        DWORD errval  = GetLastError();
+        gchar *errstr = g_win32_error_message(errval);
+        g_warning("failed to clip cursor (%ld) %s", errval, errstr);
+    }
+
+    return false;
+}
+#endif
+
 static GdkGrabStatus do_pointer_grab(SpiceDisplay *display)
 {
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
     GdkWindow *window = GDK_WINDOW(gtk_widget_get_window(GTK_WIDGET(display)));
-    GdkGrabStatus status;
+    GdkGrabStatus status = GDK_GRAB_BROKEN;
     GdkCursor *blank = get_blank_cursor();
 
     if (!gtk_widget_get_realized(GTK_WIDGET(display)))
-        return GDK_GRAB_BROKEN;
-    try_keyboard_grab(display);
+        goto end;
 
+#ifdef WIN32
+    if (!win32_clip_cursor())
+        goto end;
+#endif
+
+    try_keyboard_grab(display);
     /*
      * from gtk-vnc:
      * For relative mouse to work correctly when grabbed we need to
@@ -725,34 +814,11 @@ static GdkGrabStatus do_pointer_grab(SpiceDisplay *display)
         d->mouse_grab_active = true;
         g_signal_emit(display, signals[SPICE_DISPLAY_MOUSE_GRAB], 0, true);
     }
-#ifdef WIN32
-    {
-        RECT client_rect;
-        POINT origin;
 
-        origin.x = origin.y = 0;
-        ClientToScreen(focus_window, &origin);
-        GetClientRect(focus_window, &client_rect);
-        OffsetRect(&client_rect, origin.x, origin.y);
-        ClipCursor(&client_rect);
-    }
-#endif
-
-#ifdef GDK_WINDOWING_X11
-    if (status == GDK_GRAB_SUCCESS) {
-        int accel_numerator;
-        int accel_denominator;
-        int threshold;
-        GdkWindow *w = GDK_WINDOW(gtk_widget_get_window(GTK_WIDGET(display)));
-        Display *x_display = GDK_WINDOW_XDISPLAY(w);
-
-        XGetPointerControl(x_display, &accel_numerator, &accel_denominator,
-                           &threshold);
-        XChangePointerControl(x_display, False, False, accel_numerator,
-                              accel_denominator, threshold);
-    }
-#endif
+    if (status == GDK_GRAB_SUCCESS)
+        set_mouse_accel(display, FALSE);
 
+end:
     gdk_cursor_unref(blank);
     return status;
 }
@@ -811,10 +877,21 @@ static void try_mouse_grab(SpiceDisplay *display)
 static void mouse_wrap(SpiceDisplay *display, GdkEventMotion *motion)
 {
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
-    GdkScreen *screen = gtk_widget_get_screen(GTK_WIDGET(display));
+    gint xr, yr;
 
-    gint xr = gdk_screen_get_width(screen) / 2;
-    gint yr = gdk_screen_get_height(screen) / 2;
+#ifdef WIN32
+    RECT clip;
+    g_return_if_fail(GetClipCursor(&clip));
+    xr = clip.left + (clip.right - clip.left) / 2;
+    yr = clip.top + (clip.bottom - clip.top) / 2;
+    /* the clip rectangle has no offset, so we can't use gdk_wrap_pointer */
+    SetCursorPos(xr, yr);
+    d->mouse_last_x = -1;
+    d->mouse_last_y = -1;
+#else
+    GdkScreen *screen = gtk_widget_get_screen(GTK_WIDGET(display));
+    xr = gdk_screen_get_width(screen) / 2;
+    yr = gdk_screen_get_height(screen) / 2;
 
     if (xr != (gint)motion->x_root || yr != (gint)motion->y_root) {
         /* FIXME: we try our best to ignore that next pointer move event.. */
@@ -825,6 +902,8 @@ static void mouse_wrap(SpiceDisplay *display, GdkEventMotion *motion)
         d->mouse_last_x = -1;
         d->mouse_last_y = -1;
     }
+#endif
+
 }
 
 static void try_mouse_ungrab(SpiceDisplay *display)
@@ -839,20 +918,7 @@ static void try_mouse_ungrab(SpiceDisplay *display)
 #ifdef WIN32
     ClipCursor(NULL);
 #endif
-#ifdef GDK_WINDOWING_X11
-    {
-        int accel_numerator;
-        int accel_denominator;
-        int threshold;
-        GdkWindow *w = GDK_WINDOW(gtk_widget_get_window(GTK_WIDGET(display)));
-        Display *x_display = GDK_WINDOW_XDISPLAY(w);
-
-        XGetPointerControl(x_display, &accel_numerator, &accel_denominator,
-                           &threshold);
-        XChangePointerControl(x_display, True, True, accel_numerator,
-                              accel_denominator, threshold);
-    }
-#endif
+    set_mouse_accel(display, TRUE);
 
     d->mouse_grab_active = false;
 
@@ -1037,6 +1103,8 @@ static void send_key(SpiceDisplay *display, int scancode, SendKeyType type, gboo
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
     uint32_t i, b, m;
 
+    g_return_if_fail(scancode != 0);
+
     if (!d->inputs)
         return;
 
@@ -1133,13 +1201,24 @@ static gboolean check_for_grab_key(SpiceDisplay *display, int type, int keyval)
     return FALSE;
 }
 
+static void update_display(SpiceDisplay *display)
+{
+#ifdef WIN32
+    win32_window = display ? GDK_WINDOW_HWND(gtk_widget_get_window(GTK_WIDGET(display))) : NULL;
+#endif
+}
+
 static gboolean key_event(GtkWidget *widget, GdkEventKey *key)
 {
     SpiceDisplay *display = SPICE_DISPLAY(widget);
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
     int scancode;
 
-
+#ifdef WIN32
+    /* on windows, we ought to ignore the reserved key event? */
+    if (key->hardware_keycode == 0xff)
+        return false;
+#endif
     SPICE_DEBUG("%s %s: keycode: %d  state: %d  group %d modifier %d",
             __FUNCTION__, key->type == GDK_KEY_PRESS ? "press" : "release",
             key->hardware_keycode, key->state, key->group, key->is_modifier);
@@ -1165,6 +1244,12 @@ static gboolean key_event(GtkWidget *widget, GdkEventKey *key)
 
     scancode = vnc_display_keymap_gdk2xtkbd(d->keycode_map, d->keycode_maplen,
                                             key->hardware_keycode);
+#ifdef WIN32
+    /* MapVirtualKey doesn't return scancode with needed higher byte */
+    scancode = MapVirtualKey(key->hardware_keycode, MAPVK_VK_TO_VSC) |
+        (scancode & 0xff00);
+#endif
+
     switch (key->type) {
     case GDK_KEY_PRESS:
         send_key(display, scancode, SEND_KEY_PRESS, !key->is_modifier);
@@ -1192,6 +1277,9 @@ static guint get_scancode_from_keyval(SpiceDisplay *display, guint keyval)
         /* FIXME what about levels? */
         keycode = keys[0].keycode;
         g_free(keys);
+    } else {
+        g_warning("could not lookup keyval %u, please report a bug", keyval);
+        return 0;
     }
 
     return vnc_display_keymap_gdk2xtkbd(d->keycode_map, d->keycode_maplen, keycode);
@@ -1224,8 +1312,11 @@ static gboolean enter_event(GtkWidget *widget, GdkEventCrossing *crossing G_GNUC
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
 
     SPICE_DEBUG("%s", __FUNCTION__);
+
     d->mouse_have_pointer = true;
     try_keyboard_grab(display);
+    update_display(display);
+
     return true;
 }
 
@@ -1263,10 +1354,8 @@ static gboolean focus_in_event(GtkWidget *widget, GdkEventFocus *focus G_GNUC_UN
     sync_keyboard_lock_modifiers(display);
     update_keyboard_focus(display, true);
     try_keyboard_grab(display);
-#ifdef WIN32
-    focus_window = GDK_WINDOW_HWND(gtk_widget_get_window(widget));
-    g_return_val_if_fail(focus_window != NULL, true);
-#endif
+    update_display(display);
+
     return true;
 }
 
@@ -1276,6 +1365,7 @@ static gboolean focus_out_event(GtkWidget *widget, GdkEventFocus *focus G_GNUC_U
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
 
     SPICE_DEBUG("%s", __FUNCTION__);
+    update_display(NULL);
 
     /*
      * Ignore focus out after a keyboard grab
@@ -1389,7 +1479,8 @@ static gboolean motion_event(GtkWidget *widget, GdkEventMotion *motion)
 
             d->mouse_last_x = x;
             d->mouse_last_y = y;
-            mouse_wrap(display, motion);
+            if (dx != 0 || dy != 0)
+                mouse_wrap(display, motion);
         }
         break;
     default:
@@ -1500,14 +1591,29 @@ static gboolean configure_event(GtkWidget *widget, GdkEventConfigure *conf)
         d->wh = conf->height;
         recalc_geometry(widget);
     }
+
+    try_mouse_ungrab(display);
+    try_mouse_grab(display);
+
     return true;
 }
 
+static void update_image(SpiceDisplay *display)
+{
+    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+
+    spicex_image_create(display);
+    if (d->convert)
+        do_color_convert(display, &d->area);
+}
+
 static void realize(GtkWidget *widget)
 {
+    SpiceDisplay *display = SPICE_DISPLAY(widget);
+
     GTK_WIDGET_CLASS(spice_display_parent_class)->realize(widget);
 
-    spicex_image_create(SPICE_DISPLAY(widget));
+    update_image(display);
 }
 
 static void unrealize(GtkWidget *widget)
@@ -1818,13 +1924,14 @@ static void update_mouse_mode(SpiceChannel *channel, gpointer data)
         break;
     case SPICE_MOUSE_MODE_SERVER:
         try_mouse_grab(display);
+        d->mouse_guest_x = -1;
+        d->mouse_guest_y = -1;
         break;
     default:
         g_warn_if_reached();
     }
 
     update_mouse_pointer(display);
-    cursor_invalidate(display);
 }
 
 static void update_area(SpiceDisplay *display,
@@ -1856,7 +1963,8 @@ static void update_area(SpiceDisplay *display,
     spicex_image_destroy(display);
     d->area = area;
     if (gtk_widget_get_realized(GTK_WIDGET(display)))
-        spicex_image_create(display);
+        update_image(display);
+
     update_size_request(display);
 
     set_monitor_ready(display, true);
@@ -2447,6 +2555,41 @@ static void spice_set_keyboard_lock_modifiers(SpiceDisplay *display, uint32_t mo
     set_keyboard_led(x_display, NUM_LOCK_LED, !!(modifiers & SPICE_INPUTS_NUM_LOCK));
     set_keyboard_led(x_display, SCROLL_LOCK_LED, !!(modifiers & SPICE_INPUTS_SCROLL_LOCK));
 }
+#elif defined (WIN32)
+static guint32 get_keyboard_lock_modifiers(void)
+{
+    guint32 modifiers = 0;
+
+    if (GetKeyState(VK_CAPITAL) & 1) {
+        modifiers |= SPICE_INPUTS_CAPS_LOCK;
+    }
+    if (GetKeyState(VK_NUMLOCK) & 1) {
+        modifiers |= SPICE_INPUTS_NUM_LOCK;
+    }
+    if (GetKeyState(VK_SCROLL) & 1) {
+        modifiers |= SPICE_INPUTS_SCROLL_LOCK;
+    }
+
+    return modifiers;
+}
+
+static void sync_keyboard_lock_modifiers(SpiceDisplay *display)
+{
+    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+    guint32 modifiers;
+    GdkWindow *w;
+
+    if (d->disable_inputs)
+        return;
+
+    w = gtk_widget_get_parent_window(GTK_WIDGET(display));
+    if (w == NULL) /* it can happen if the display is not yet shown */
+        return;
+
+    modifiers = get_keyboard_lock_modifiers();
+    if (d->inputs)
+        spice_inputs_set_key_locks(d->inputs, modifiers);
+}
 #else
 static void sync_keyboard_lock_modifiers(SpiceDisplay *display)
 {
