@@ -81,6 +81,7 @@ enum {
     PROP_SESSION,
     PROP_AUTO_CONNECT,
     PROP_AUTO_CONNECT_FILTER,
+    PROP_REDIRECT_ON_CONNECT,
 };
 
 enum
@@ -96,6 +97,7 @@ struct _SpiceUsbDeviceManagerPrivate {
     SpiceSession *session;
     gboolean auto_connect;
     gchar *auto_connect_filter;
+    gchar *redirect_on_connect;
 #ifdef USE_USBREDIR
     libusb_context *context;
     GUdevClient *udev;
@@ -104,7 +106,12 @@ struct _SpiceUsbDeviceManagerPrivate {
     gboolean event_thread_run;
     libusb_device **coldplug_list; /* Avoid needless reprobing during init */
     struct usbredirfilter_rule *auto_conn_filter_rules;
+    struct usbredirfilter_rule *redirect_on_connect_rules;
     int auto_conn_filter_rules_count;
+    int redirect_on_connect_rules_count;
+#ifdef G_OS_WIN32
+    SpiceWinUsbDriver     *installer;
+#endif
 #endif
     GPtrArray *devices;
     GPtrArray *channels;
@@ -143,6 +150,8 @@ static void spice_usb_device_manager_uevent_cb(GUdevClient     *client,
                                                gpointer         user_data);
 static void spice_usb_device_manager_add_dev(SpiceUsbDeviceManager  *self,
                                              GUdevDevice            *udev);
+static void spice_usb_device_manager_check_redir_on_connect(
+    SpiceUsbDeviceManager *self, SpiceChannel *channel);
 
 static SpiceUsbDeviceInfo *spice_usb_device_new(libusb_device *libdev);
 static SpiceUsbDevice *spice_usb_device_ref(SpiceUsbDevice *device);
@@ -155,9 +164,6 @@ static void  spice_usb_device_set_state(SpiceUsbDevice *device, guint8 s);
 
 static gboolean spice_usb_device_equal_libdev(SpiceUsbDevice *device,
                                               libusb_device *libdev);
-static SpiceUsbDevice *
-spice_usb_device_manager_libdev_to_device(SpiceUsbDeviceManager *self,
-                                          libusb_device *libdev);
 static libusb_device *
 spice_usb_device_manager_device_to_libdev(SpiceUsbDeviceManager *self,
                                           SpiceUsbDevice *device);
@@ -248,17 +254,6 @@ static gboolean spice_usb_device_manager_initable_init(GInitable  *initable,
         return FALSE;
     }
 
-    /* Start listening for usb channels connect/disconnect */
-    g_signal_connect(priv->session, "channel-new",
-                     G_CALLBACK(channel_new), self);
-    g_signal_connect(priv->session, "channel-destroy",
-                     G_CALLBACK(channel_destroy), self);
-    list = spice_session_get_channels(priv->session);
-    for (it = g_list_first(list); it != NULL; it = g_list_next(it)) {
-        channel_new(priv->session, it->data, (gpointer*)self);
-    }
-    g_list_free(list);
-
     /* Start listening for usb devices plug / unplug */
     priv->udev = g_udev_client_new(subsystems);
     g_signal_connect(G_OBJECT(priv->udev), "uevent",
@@ -275,6 +270,17 @@ static gboolean spice_usb_device_manager_initable_init(GInitable  *initable,
     libusb_free_device_list(priv->coldplug_list, 1);
     priv->coldplug_list = NULL;
 
+    /* Start listening for usb channels connect/disconnect */
+    g_signal_connect(priv->session, "channel-new",
+                     G_CALLBACK(channel_new), self);
+    g_signal_connect(priv->session, "channel-destroy",
+                     G_CALLBACK(channel_destroy), self);
+    list = spice_session_get_channels(priv->session);
+    for (it = g_list_first(list); it != NULL; it = g_list_next(it)) {
+        channel_new(priv->session, it->data, (gpointer*)self);
+    }
+    g_list_free(list);
+
     return TRUE;
 #else
     g_set_error_literal(err, SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
@@ -299,6 +305,10 @@ static void spice_usb_device_manager_finalize(GObject *gobject)
     if (priv->event_thread)
         g_thread_join(priv->event_thread);
     free(priv->auto_conn_filter_rules);
+#ifdef G_OS_WIN32
+    if (priv->installer)
+        g_object_unref(priv->installer);
+#endif
 #endif
 
     g_free(priv->auto_connect_filter);
@@ -331,6 +341,9 @@ static void spice_usb_device_manager_get_property(GObject     *gobject,
     case PROP_AUTO_CONNECT_FILTER:
         g_value_set_string(value, priv->auto_connect_filter);
         break;
+    case PROP_REDIRECT_ON_CONNECT:
+        g_value_set_string(value, priv->redirect_on_connect);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
         break;
@@ -374,6 +387,30 @@ static void spice_usb_device_manager_set_property(GObject       *gobject,
         priv->auto_connect_filter = g_strdup(filter);
         break;
     }
+    case PROP_REDIRECT_ON_CONNECT: {
+        const gchar *filter = g_value_get_string(value);
+#ifdef USE_USBREDIR
+        struct usbredirfilter_rule *rules = NULL;
+        int r = 0, count = 0;
+
+        if (filter)
+            r = usbredirfilter_string_to_rules(filter, ",", "|",
+                                               &rules, &count);
+        if (r) {
+            if (r == -ENOMEM)
+                g_error("Failed to allocate memory for redirect-on-connect");
+            g_warning("Error parsing redirect-on-connect string, keeping old filter\n");
+            break;
+        }
+
+        free(priv->redirect_on_connect_rules);
+        priv->redirect_on_connect_rules = rules;
+        priv->redirect_on_connect_rules_count = count;
+#endif
+        g_free(priv->redirect_on_connect);
+        priv->redirect_on_connect = g_strdup(filter);
+        break;
+    }
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
         break;
@@ -448,6 +485,21 @@ static void spice_usb_device_manager_class_init(SpiceUsbDeviceManagerClass *klas
     g_object_class_install_property(gobject_class, PROP_AUTO_CONNECT_FILTER,
                                     pspec);
 
+    /**
+     * SpiceUsbDeviceManager:redirect-on-connect:
+     *
+     * Set a string specifying a filter selecting USB devices to automatically
+     * redirect after a Spice connection has been established.
+     *
+     * See SpiceUsbDeviceManager:auto-connect-filter: for the filter string
+     * format.
+     */
+    pspec = g_param_spec_string("redirect-on-connect", "Redirect on connect",
+               "Filter selecting USB devices to redirect on connect", NULL,
+               G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+    g_object_class_install_property(gobject_class, PROP_REDIRECT_ON_CONNECT,
+                                    pspec);
+
     /**
      * SpiceUsbDeviceManager::device-added:
      * @manager: the #SpiceUsbDeviceManager that emitted the signal
@@ -568,6 +620,8 @@ static void channel_new(SpiceSession *session, SpiceChannel *channel,
                                            self->priv->context);
         spice_channel_connect(channel);
         g_ptr_array_add(self->priv->channels, channel);
+
+        spice_usb_device_manager_check_redir_on_connect(self, channel);
     }
 }
 
@@ -829,7 +883,6 @@ static void spice_usb_device_manager_drv_install_cb(GObject *gobject,
 
     status = spice_win_usb_driver_install_finish(installer, res, &err);
 
-    g_object_unref(installer);
     spice_usb_device_unref(device);
 
     spice_usb_device_set_state(device, SPICE_USB_DEVICE_STATE_NONE);
@@ -868,7 +921,7 @@ static gpointer spice_usb_device_manager_usb_ev_thread(gpointer user_data)
 
     while (priv->event_thread_run) {
         rc = libusb_handle_events(priv->context);
-        if (rc) {
+        if (rc && rc != LIBUSB_ERROR_INTERRUPTED) {
             const char *desc = spice_usbutil_libusb_strerror(rc);
             g_warning("Error handling USB events: %s [%i]", desc, rc);
         }
@@ -919,15 +972,54 @@ void spice_usb_device_manager_stop_event_listening(
         priv->event_thread_run = FALSE;
 }
 
-void spice_usb_device_manager_device_error(
-    SpiceUsbDeviceManager *self, libusb_device *libdev, GError *err)
+static void spice_usb_device_manager_check_redir_on_connect(
+    SpiceUsbDeviceManager *self, SpiceChannel *channel)
 {
+    SpiceUsbDeviceManagerPrivate *priv = self->priv;
+    GSimpleAsyncResult *result;
     SpiceUsbDevice *device;
+    libusb_device *libdev;
+    guint i;
 
+    if (priv->redirect_on_connect == NULL)
+        return;
+
+    for (i = 0; i < priv->devices->len; i++) {
+        device = g_ptr_array_index(priv->devices, i);
+
+        if (spice_usb_device_manager_is_device_connected(self, device))
+            continue;
+
+        libdev = spice_usb_device_manager_device_to_libdev(self, device);
+
+        if (usbredirhost_check_device_filter(
+                            priv->redirect_on_connect_rules,
+                            priv->redirect_on_connect_rules_count,
+                            libdev, 0) == 0) {
+            /* Note: re-uses spice_usb_device_manager_connect_device_async's
+               completion handling code! */
+            result = g_simple_async_result_new(G_OBJECT(self),
+                               spice_usb_device_manager_auto_connect_cb,
+                               spice_usb_device_ref(device),
+                               spice_usb_device_manager_connect_device_async);
+            spice_usbredir_channel_connect_device_async(
+                               SPICE_USBREDIR_CHANNEL(channel),
+                               libdev, device, NULL,
+                               spice_usb_device_manager_channel_connect_cb,
+                               result);
+            libusb_unref_device(libdev);
+            return; /* We've taken the channel! */
+        }
+
+        libusb_unref_device(libdev);
+    }
+}
+
+void spice_usb_device_manager_device_error(
+    SpiceUsbDeviceManager *self, SpiceUsbDevice *device, GError *err)
+{
     g_return_if_fail(SPICE_IS_USB_DEVICE_MANAGER(self));
-    g_return_if_fail(libdev != 0);
-
-    device = spice_usb_device_manager_libdev_to_device(self, libdev);
+    g_return_if_fail(device != NULL);
 
     g_signal_emit(self, signals[DEVICE_ERROR], 0, device, err);
 }
@@ -1091,6 +1183,7 @@ _spice_usb_device_manager_connect_device_async(SpiceUsbDeviceManager *self,
         }
         spice_usbredir_channel_connect_device_async(channel,
                                  libdev,
+                                 device,
                                  cancellable,
                                  spice_usb_device_manager_channel_connect_cb,
                                  result);
@@ -1122,7 +1215,10 @@ void spice_usb_device_manager_connect_device_async(SpiceUsbDeviceManager *self,
     UsbInstallCbInfo *cbinfo;
 
     spice_usb_device_set_state(device, SPICE_USB_DEVICE_STATE_INSTALLING);
-    installer = spice_win_usb_driver_new();
+    if (! self->priv->installer) {
+        self->priv->installer = spice_win_usb_driver_new();
+    }
+    installer = self->priv->installer;
     cbinfo = g_new0(UsbInstallCbInfo, 1);
     cbinfo->manager     = self;
     cbinfo->device      = spice_usb_device_ref(device);
@@ -1186,9 +1282,13 @@ void spice_usb_device_manager_disconnect_device(SpiceUsbDeviceManager *self,
     UsbInstallCbInfo *cbinfo;
 
     g_warn_if_fail(device != NULL);
+    g_warn_if_fail(self->priv->installer != NULL);
 
     spice_usb_device_set_state(device, SPICE_USB_DEVICE_STATE_UNINSTALLING);
-    installer = spice_win_usb_driver_new();
+    if (! self->priv->installer) {
+        self->priv->installer = spice_win_usb_driver_new();
+    }
+    installer = self->priv->installer;
     cbinfo = g_new0(UsbInstallCbInfo, 1);
     cbinfo->manager     = self;
     cbinfo->device      = spice_usb_device_ref(device);
@@ -1471,18 +1571,6 @@ spice_usb_device_equal_libdev(SpiceUsbDevice *device,
     return ((bus1 == bus2) && (addr1 == addr2));
 }
 
-static SpiceUsbDevice *
-spice_usb_device_manager_libdev_to_device(SpiceUsbDeviceManager *self,
-                                          libusb_device *libdev)
-{
-    guint8 bus, addr;
-
-    bus  = libusb_get_bus_number(libdev);
-    addr = libusb_get_device_address(libdev);
-
-    return spice_usb_device_manager_find_device(self, bus, addr);
-}
-
 /*
  * Caller must libusb_unref_device the libusb_device returned by this function.
  * Returns a libusb_device, or NULL upon failure
